#
# File: generateConfigBackup.pl
# Description: This script generates a backup of all of the configuration
#           files.  This may include a configuration database on primaries.
#
#
require "$ENV{FT_DIR}/bin/product.pl";
require "$ENV{FT_DIR}/bin/ftcutils.pl";

my $target_dir = "/tmp/configbackup";

{

my ($ft_dir, $osType);
my $errorFile = "$ENV{$_LOG_ENV_VAR_}/configbackup.log";
my $persistedConfigDir = $ENV{$_PERSISTED_CONFIG_ENV_VAR_};

open (CBERR, ">$errorFile"); # Don't append to keep log small

if ($#ARGV+1 > 1) {
    print CBERR "Invalid number of arguments.  Call syntax = generateConfigBackup [-now]\n";
    exit(-1);
}

my $immediate = TRUE if ($ARGV[0] eq "-now");

#===========================================================================
# Make sure the $_DIR_ENV_VAR_ environment variable is defined
#===========================================================================
if (! $ENV{$_DIR_ENV_VAR_}) 
{
    print CBERR "The environment variable: \"$_DIR_ENV_VAR_\" must be set\n";
    exit(-1);
}

# Get os type
$osType = `uname -s`;
chop($osType); 

if ($osType !~ "VMkernel") {
    # For now, don't do anything on ESX
    print CBERR "No backup needed for ESX\n";
    # Set the status for the ft_init.pl script
    print "OK\n";
    exit(0);
}

# set up the source directory of the backup
my $configSourceDir = $ENV{$_CONFIG_ENV_VAR_};

my $myNode = &get_local_hostname();
if ($myNode eq "") {
    $myNode = "localhost";
    print CBERR "Cannot identify host, make sure host name is configured.";
    print CBERR "Choosing $myNode ...\n";
}

$DomainName = $ENV{FT_DOMAIN};

my $dbSubDir = "${DomainName}_$myNode";
my $dbBackupSourceDir = "$ENV{$_CONFIG_ENV_VAR_}/configbackup";

`/bin/mkdir -p $target_dir` unless ( -e $target_dir);
`/bin/mkdir -p $persistedConfigDir` unless ( -e $persistedConfigDir);

# clean up any residue from prior backups
`/bin/rm -rf ${target_dir}/*`;

# copy the the configuration files.  If this is a primary, the database
# will also be in that directory
print CBERR "Copying $configSourceDir to $target_dir\n";
my $result = `/bin/cp -r $configSourceDir/* $target_dir`;
if ($?) {
    print CBERR "Error copying $configSourceDir to $target_dir: $result\n";
    &cleanup_and_exit(1);
}

if (-e $dbBackupSourceDir) {
    # Copy the backed up crash-consistent database directory if it exists
    print CBERR "Copying the backed up database directory $dbBackupSourceDir\n";
    $result = `cp -r ${dbBackupSourceDir}/${dbSubDir} $target_dir`;
    if ($?) {
        print CBERR "Error copying $configSourceDir to $target_dir: $result\n";
        &cleanup_and_exit(1);
    }
}


# Preserve any Isis incarnation files - these will not be present on secondaries
print CBERR "Backing up Isis incarnation files\n";
$result = `/bin/cp -f $ENV{$_LOG_ENV_VAR_}/backbone/*.incarn $target_dir 2>&1`;
if ($?) {
    print CBERR "Warning:Error copying incarnation files to $target_dir: $result\n";
    close CBERR;
    &cleanup_and_exit(1);
}

my $configBackupTar = "configBackup.tar.gz";
print CBERR "creating the zipped tar file $configBackupTar\n";
chdir($target_dir);

$result = `/bin/tar czf $configBackupTar * 2>&1`;
if ($?) {
    print CBERR "Error tarring ${target_dir} to $configBackupTar: $result\n";
    close CBERR;
    &cleanup_and_exit(1);
}

# First copy the file to the stable storage under a temporary 
# name and then rename it.  This way a failure during the copy 
# will leave the old saved config intact 
my $cmd = "/bin/mv -f $configBackupTar $persistedConfigDir/$configBackupTar.tmp";
print CBERR "About to $cmd\n";
$result = `$cmd 2>&1`;
if ($?) {
    print CBERR "Error $cmd failed: $result\n ";
    close CBERR;
    &cleanup_and_exit(1);
}

$cmd = "/bin/mv -f $persistedConfigDir/$configBackupTar.tmp " .
    "$persistedConfigDir/$configBackupTar";
print CBERR "About to $cmd\n";
$result = `$cmd 2>&1`;
if ($?) {
    print CBERR "Error $cmd failed: $result";
    close CBERR;
    &cleanup_and_exit(1);
}

# Critical state to back up now
if ($immediate) {
    print CBERR "Immediate Backup occurring\n";
    $cmd = "/sbin/backup.sh 0";
    $result = `$cmd 2>&1`;
    my $returnCode = $?;
    if ($returnCode) {
        print CBERR "Error /sbin/backup.sh failed: $returnCode";
        close CBERR;
        &cleanup_and_exit(1);
    }
    else {
    }
}
else {
    # Backup the config backup to the local.tgz file for
    # recovery after failures
    # trigger the USB write if within the scheduled budget 
    $result = `/bin/vim-cmd hostsvc/firmware/sync_config 2>&1`;
    my $returnCode = $?;
    if ($returnCode) {
        print CBERR "Error sync_config failed: $returnCode";
        close CBERR;
        &cleanup_and_exit(1);
    }
}

print CBERR "Cleaning up temporary working directory\n";
close CBERR;

# Set the status for the ft_init.pl script
print "OK\n";
&cleanup_and_exit(0);
}

sub cleanup_and_exit {
    my $exit_code = shift;
    `/bin/rm -rf ${target_dir}`;
    exit($exit_code);
}
    
