#!/bin/sh
# Copyright 2008 VMware Inc.,

if [ -f /etc/vmware/sysboot.conf ] ; then
   . /etc/vmware/sysboot.conf
fi

: "${VERBOSE:=no}"
: "${SYSBOOT_TIME:=no}"
: "${SYSBOOT_QUIET:=no}"
: "${SYSBOOT_DEBUG:=no}"
: "${SYSBOOT_COLOR:=yes}"
: "${LOGFILE:=/messages.log}"

#
# Error Code Enumeration
#
SUCCESS=0
NON_CRITICAL_FAILURE=1
FAILURE=2
FATAL_ERROR=3
HALT=4

TRUE=0
FALSE=1

INDENTATION=""

if [ "${SYSBOOT_COLOR}" = "yes" ] ; then
   GOOD='[01;32m'
   WARN='[01;33m'
   ERROR='[01;31m'
   NORMAL='[00;00m'
   HIGHLIGHT='[01;36m'
   BRACKET='[01;34m'
else
   unset GOOD WARN ERROR NORMAL HIGHLIGHT BRACKET
fi

COLS=$(stty size 2> /dev/null | awk '{ print $2 }')
RIGHT_ALIGN='[A['$(( ${COLS:-80} - 9 ))'C'

### echo-functions ###

Indent()
{
   if [ ${#INDENTATION} -ge 40 ] ; then
      return
   fi

   INDENTATION="${INDENTATION}   "
}

Outdent()
{
   if [ ${#INDENTATION} -eq 0 ] ; then
      return
   fi

   INDENTATION="${INDENTATION%   }"
}

Log()
{
   local tag=

   if [ $# -gt 1 ] ; then
      tag="$1"
      shift
   else
      tag="sysboot"
   fi

   if [ "${SYSBOOT_DEBUG}" != "no" ] ; then
      echo "${tag}: $*" >&2
   fi

   echo "${tag}: $*" >> "${LOGFILE}"

   if [ -f /proc/vmware/log ] ; then
      echo -n "${tag}: ${*}" >> /proc/vmware/log
   elif [ -e /dev/klog ]; then
      # truncate log msg to fit on a klog line
      echo "${tag}: ${*}" | cut -c1-126 > /dev/klog
   fi
}

Info()
{
   Log "sysboot" "$*"

   if [ "${SYSBOOT_QUIET}" != "yes" ] ; then
      echo " ${INDENTATION}${GOOD}*${NORMAL} $*"
   fi
}

Warn()
{
   Log "sysboot" "$*"
   echo " ${INDENTATION}${WARN}*${NORMAL} $*"
}

Error()
{
   local width=$(( ${COLS:-80} - ${#INDENTATION} - 11 ))

   Log "sysboot" "${*}"

   echo ${*} | sed -e :a -e '/^.\{'${width}'\}/!{ $q; P; bb; }'                                            \
                   -e '/^.\{'${width}'\}/{ s/.\{'${width}'\}/&\n/; s/^\(.*\) \(.*\)\n/\1\n\2/; P; bb; }'   \
                   -e :b -e 's/^.*\n//' -e ba | while read line ; do
      echo " ${INDENTATION}${ERROR}*${NORMAL} ${line}"
   done
}

Begin()
{
   if [ "${SYSBOOT_QUIET}" != "yes" -o "${SYSBOOT_TIME}" = "yes" ] ; then
      Info "$* ..."
      Indent
   fi
}

End()
{
   local code="${1:-0}" msg=

   if [ "${SYSBOOT_QUIET}" != "yes" ] ; then
      if [ "${code}" = "${SUCCESS}" ] ; then
         msg="${BRACKET}[ ${GOOD}ok${BRACKET} ]${NORMAL}"
      else
         msg="${BRACKET}[ ${ERROR}!!${BRACKET} ]${NORMAL}"
      fi

      echo "${RIGHT_ALIGN}  ${msg}"
      Outdent
   fi
}


### module-functions ###

_get_module_name()
{
   local alias=

   while [ ${#} -gt 0 ] ; do
      case "${1}" in
         -m) shift ; alias="${1}" ; shift ;;
         -?) shift ;;
         *) break ;;
      esac
   done

   echo "${1}$([ -n "${alias}" ] && echo " as ${alias}")"
}

Insmod()
{
   if [ "${VERBOSE}" != "no" ] ; then
      Info "Loading Linux Module '" "$1" "'"
   else
      Log "Loading Linux Module '" "$1" "'"
   fi

   insmod "$@" >> "${LOGFILE}" 2>&1
   return $?
}

VmkloadMod()
{
   if [ "${VERBOSE}" != "no" ] ; then
      Info "Loading VMKernel Module '"$(_get_module_name "$@")"'"
   else
      Log "Loading VMkernel Module '"$(_get_module_name "$@")"'"
   fi

   esxcfg-module -v 3 "$@" >> "${LOGFILE}" 2>&1
   return $?
}

#
# UnloadModule --
#
# Results:
#       TRUE if the module was unloaded, FALSE otherwise
#
# Side Effects:
#       Specified module is unloaded.  If the module does not immediately
#       unload, the time taken to unload is logged.
#
# NOTE: THIS IS AN EXTREMELY UNSAFE OPERATION.  DO NOT USE!
#
UnloadModule()
{
   local ModuleUnloaded=0
   local ModuleUnloadWarningFactor=8

   # Some modules refuse to unload right away as they work still queued.  Back
   # off incrementally, but we dont want to wait for ever, so cap at 32 seconds.
   for delay in 0 1 2 4 8 16 32 ; do
      sleep ${delay}

      VmkloadMod -u ${1} && {
         ModuleUnloaded=1

         if [ ${delay} -gt ${ModuleUnloadWarningFactor} ] ; then
            Log "Took $(( ${delay} * 2 - 1 )) seconds"
         fi

         break
      }
   done

   if [ ${ModuleUnloaded} -eq 1 ] ; then
      Log "Module '${1}' unloaded successfully"
      return ${TRUE}
   else
      Log "Failed to unload module '${1}'"
      return ${FALSE}
   fi
}

IsModuleLoaded()
{
   if [ $(vmkload_mod -bl | grep -c "$1") -eq 0 ] ; then
      return ${FALSE}
   else
      return ${TRUE}
   fi
}


### configuration-functions ###

BootOptionSpecified()
{
   local cmdline=

   Log "Checking if the '${1}' option is specified on the kernel boot line"

   if HostType_OSIsPureVMK ; then
      cmdline=$(vsish -d -e get /system/bootCmdLine)
   else
      cmdline=$(cat /proc/cmdline)
   fi

   if echo ${cmdline} | grep -q "${1}" 2> /dev/null ; then
      return ${TRUE}
   fi

   return ${FALSE}
}

GetBootOption()
{
   local value="$(echo "$1" | sed -e 's:\([\^\$\.\*\/]\):\\\1:g')"
   local cmdline=

   Log "Getting '${1}' parameter from kernel boot line"

   if HostType_OSIsPureVMK ; then
      cmdline=$(vsish -d -e get /system/bootCmdLine)
   else
      cmdline=$(cat /proc/cmdline)
   fi

   echo ${cmdline} | sed -n -e 's:.*'"${value}"'=\(\([^ \\]\|\\.\)\+\).*:\1:p;s/\\\(.\)/\1/gp' 2> /dev/null
}

GetRawConfigOption()
{
   Log "Getting '$1' parameter from esx.conf"
   echo "$(esxcfg-advcfg -c "${1}")"
}

GetVMKernelOption()
{
   Log "Getting '$1' parameter from esx.conf"
   echo "$(esxcfg-advcfg -q -g "${1}")"
}

SetVMKernelOption()
{
   Log "Setting '$1' to '$2'"
   Exec esxcfg-advcfg -s "${2}" "${1}"
}

### utility-functions ###

HostType_OSIsPureVMK()
{
   if [ "$(uname -s)" = "VMkernel" ] ; then
      return ${TRUE}
   fi

   return ${FALSE}
}

Panic()
{
   Error "$*"
   Exec esxcfg-init --alert "Error: $*"
   Exec chvt 1
}

USB20Present()
{
   if [ $(vmkload_mod -bl | grep -c ehci-hcd) -eq 0 ] ; then
      return ${FALSE}
   else
      return ${TRUE}
   fi
}

Exec()
{
   if [ "${VERBOSE}" != "no" ] ; then
      Info "Executing '$*'"
   else
      Log "Executing '$*'"
   fi

   "$@" >> "${LOGFILE}" 2>&1
   return $?
}

GetPIDsForProcess()
{
   local ProcessName="${1}"
   echo $(pidof "${ProcessName}")
}

__LinuxIsProcessRunning()
{
   local ProcessName="${1}" PIDs=

   PIDs=$(GetPIDsForProcess "${ProcessName}")

   if [ -n "${PIDs}" ] ; then
      return ${TRUE}
   fi

   return ${FALSE}
}

GetWIDsForProcess()
{
   local ProcessName="${1}"
   echo $(ps | awk '$4 == "'${ProcessName}'" { print $1 }')
}

__VMkernelIsProcessRunning()
{
   local ProcessName="${1}" WorldIDs=

   WorldIDs=$(GetWIDsForProcess "${ProcessName}")

   if [ -n "${WorldIDs}" ] ; then
      return ${TRUE}
   fi

   return ${FALSE}
}

# Use the system name to determine behaviour
alias IsProcessRunning=__$(uname -s)IsProcessRunning

IsPIDRunning()
{
   kill -0 "$1" 2> /dev/null
}

__LinuxKillProcess()
{
   local ProcessName="${1}" Signal=${2:=9}
   local retval=

   for ProcessID in $(GetPIDsForProcess "${ProcessName}") ; do
      if IsPIDRunning ${ProcessID} ; then
         if [ "${VERBOSE}" = "yes" ] ; then
            Log "Killing Process ${ProcessID} with ${Signal}"
         fi

         retval=$(kill -${Signal} ${ProcessID} >/dev/null 2>&1 ; echo ${?})

         if IsPIDRunning ${ProcessID} ; then
            if [ "${VERBOSE}" = "yes" ] ; then
               Log "Force Killing Process ${ProcessID} with -KILL"
            fi

            retval=$(kill -KILL ${ProcessID} >/dev/null 2>&1 ; echo ${?})
         fi
      fi
   done
}

IsWIDRunning()
{
   kill -0 "$1" 2> /dev/null
}

__VMkernelKillProcess()
{
   local ProcessName="${1}" Signal=${2:=9}
   local retval=

   for WorldID in $(GetWIDsForProcess "${ProcessName}") ; do
      if IsWIDRunning ${WorldID} ; then
         if [ "${VERBOSE}" = "yes" ] ; then
            Log "Killing Process ${WorldID} with ${Signal}"
         fi

         retval=$(kill -${Signal} ${WorldID} >/dev/null 2>&1 ; echo ${?})

         if IsWIDRunning ${WorldID} ; then
            if [ "${VERBOSE}" = "yes" ] ; then
               Log "Force Killing Process ${WorldID} with -KILL"
            fi

            retval=$(kill -KILL ${WorldID} >/dev/null 2>&1 ; echo ${?})
         fi
      fi
   done

   return ${retval}
}

# Use the system name to determine behaviour
alias KillProcess=__$(uname -s)KillProcess

KillUWShadowProcess()
{
   local UserworldName="${1}"
   local vmkloadapp="/usr/lib/vmware/bin/vmkload_app"

   local uwPid=$(ps x | awk '$5 == "'${vmkloadapp}'" && $6 == "'${UserworldName}'" { print $1 }')
   if [ -n "${uwPid}" ] ; then
      kill ${uwPid}
      return ${TRUE}
   fi
   return ${FALSE}
}

__LinuxRecoveryShell()
{
   Exec chvt 1
   /etc/vmware/init/bin/shellrc "$@"
}

__VMkernelRecoveryShell()
{
   Exec chvt 1
   /sbin/techsupport.sh -S 38400 tty1
}

alias RecoveryShell=__$(uname -s)RecoveryShell


### visor-functions ###

IsEmbedded()
{
   if $(esxcfg-info -e > /dev/null 2>&1) ; then
      return ${TRUE}
   else
      return ${FALSE}
   fi
}

IsUSBBoot()
{
   if [ $(bootOption -r -u 2> /dev/null) -eq 0 ] ; then
      return ${FALSE}
   else
      return ${TRUE}
   fi
}

FileSystemCheckRequired()
{
   local fsCheck=$(vsish -e get /vmkModules/vfat/fsInfo | awk -F: '/fsCheck/ {print $2}')
   if [ $(bootOption -r -f) -eq 0 -a "$fsCheck" = "0" ] ; then
      return ${FALSE}
   else
      return ${TRUE}
   fi
}

RunFileSystemCheck()
{
   local BootDevice=

   Log "Verifying default filesystems"
   Exec vmkfstools -V
   Exec sleep 5

   BootDevice=$(GetBootHBA)

   for partition in 5 6 8 ; do
      Disk="/dev/disks/${BootDevice}:${partition}"

      if [ -f "${Disk}" ] ; then
         Exec dosfsck -a -v "${Disk}" || {
            Log "Tried to repair possible corruption on ${Disk}"
            Exec esxcfg-init -a "Possible corruption on ${Disk}"
            chvt 1
         }
      fi
   done
}

GetHBAPFromVolume()
{
   echo $(vmkfstools -P "$1" 2> /dev/null | sed -n '$s/^.\(.\+:[0-9]\+\)$/\1/p')
}

GetHBAFromVolume()
{
   echo $(vmkfstools -P "$1" 2> /dev/null | sed -n '$s/^.\(.\+\):[0-9]\+$/\1/p')
}

GetVolumeSize()
{
   echo  $(vmkfstools -P "$1" 2> /dev/null | sed -n 's/^Capacity \([0-9]*\) .*/\1/p')
}

GetVolumeType()
{
   echo  $(vmkfstools -P "$1" 2> /dev/null | sed -n 's/^\([A-Za-z]*\)-.*/\1/p')
}

GetPartitionFromHBAP()
{
   echo $(echo "$1" | awk -F: '{ print $NF }')
}

GetVolumeFromHBAP()
{
   local Volume=

   for volume in /vmfs/volumes/* ; do
      if [ -h "${volume}" ] ; then
         continue
      fi

      local VolumeHBAP=$(GetHBAPFromVolume "${volume}")

      if [ "${VolumeHBAP}" = "$1" ] ; then
         Volume="${volume}"
         break
      fi
   done

   echo "${Volume}"
}

GetBootHBA()
{
   local BootVolume=$(GetPrimaryBootVolume)
   local BootHBA=

   if [ -n "${BootVolume}" ] ; then
      BootHBA=$(GetHBAFromVolume "${BootVolume}")
   fi

   echo ${BootHBA}
}

GetPrimaryBootVolume()
{
   local BootUUID=$(esxcfg-info -b 2> /dev/null)
   local BootVolume=

   if [ -n "${BootUUID}" ] ; then
      BootVolume="/vmfs/volumes/${BootUUID}"
   fi

   echo ${BootVolume}
}

GetPrimaryBootHBAP()
{
   local BootVolume=$(GetPrimaryBootVolume)
   local BootHBAP=

   if [ -n "${BootVolume}" ] ; then
      BootHBAP=$(GetHBAPFromVolume "${BootVolume}")
   fi

   echo ${BootHBAP}
}

GetSecondaryBootHBAP()
{
   local PrimaryBootBankPartition=$(GetPartitionFromHBAP "$(GetPrimaryBootHBAP)")
   local SecondaryBootHBAP=

   # We determine the secondary bootbank based upon the location of the
   # primary bootbank on disk.  If the primary bootbank is the 5th partition
   # then the secondary boot bank is the partition that immediate follows, or
   # the 6th.  In the case that the primary bootbank is not on the 5th
   # partition, we assume that it is the 4th partition (as the bootbanks are
   # located inside an extended partition and there are two primary
   # partitions), and therefore the secondary bootbank is the 5th partition.
   if [ "${PrimaryBootBankPartition}" = "5" ] ; then
      SecondaryBootHBAP="$(GetBootHBA):6"
   else
      SecondaryBootHBAP="$(GetBootHBA):5"
   fi

   echo ${SecondaryBootHBAP}
}

GetSecondaryBootVolume()
{
   echo $(GetVolumeFromHBAP "$(GetSecondaryBootHBAP)")
}

GetStoreVolume()
{
   echo $(GetVolumeFromHBAP "$(GetBootHBA):8")
}

SetupBootBanks()
{
   local PrimaryBootBank= SecondaryBootBank=

   # Rescan to discover volumes
   Exec vmkfstools -V

   PrimaryBootBank=$(GetPrimaryBootVolume)
   SecondaryBootBank=$(GetSecondaryBootVolume)

   # Detect race while detecting partitions 
   if [ "${PrimaryBootBank}" = "${SecondaryBootBank}" ] ; then
      Log "Invalid Secondary Boot Bank Dectected (matches primary)"
      return ${FALSE}
   fi

   # Primary bootbank
   if [ ! -d "${PrimaryBootBank}" ] ; then
      Log "Invalid Primary Boot Bank Detected: ${PrimaryBootBank}"
      return ${FALSE}
   fi

   Log "Setting /bootbank to ${PrimaryBootBank}"
   if [ ! -L "/bootbank" ] ; then
      Exec ln -s "${PrimaryBootBank}" /bootbank || {
         Log "Failed to link /bootbank"
         return ${FALSE}
      }
   fi

   # Secondary bootbank
   if [ ! -d "${SecondaryBootBank}" ] ; then
      Log "Invalid Secondary Boot Bank Detected: ${SecondaryBootBank}"
      return ${FALSE}
   fi

   Log "Setting /altbootbank to ${SecondaryBootBank}"
   if [ ! -L "/altbootbank" ] ; then
      Exec ln -s "${SecondaryBootBank}" /altbootbank || {
         Log "Failed to link /altbootbank"
         return ${FALSE}
      }
   fi

   return ${TRUE}
}

SetupStore()
{
   local Store="$(GetStoreVolume)"

   if [ ! -d "${Store}" ] ; then
      Log "Invalid Store Detected: ${Store}"
      return ${FALSE}
   fi

   Log "Setting /store to ${Store}"
   if [ ! -L "/store" ] ; then
      Exec ln -s "${Store}" /store || {
         Log "Failed to link /store"
         return ${FALSE}
      }
   fi

   return ${TRUE}
}

RestoreConfiguration()
{
   local Module="$1"

   if [ ! -f "${Module}" ] ; then
      return ${FALSE}
   fi

   Log "Extracting Module: ${Module}"
   Exec tar xzf "${Module}" -C / || return ${FALSE}

   export FIRST_BOOT=${FALSE}

   return ${TRUE}
}

IsTroubleShootingMode()
{
   if [ $(grep -c trouble /proc/cmdline) -eq 1 ] ; then
      return ${TRUE}
   else
      return ${FALSE}
   fi
}
